/*==========================================================================;
 *
 *  AUTHOR:     Subash, Agila
 *
 *  FILE NAME:  binary.h
 *
 *  CONTENTS:  Definition of the binary AVL tree to store ground atoms.
 *              class 'Binary', structure 'AVLNode'
 *
 *  DATE:      March 24, 2004
 ***************************************************************************/

#ifndef Binaryh
#define Binaryh

#include "common.h"

struct GroundTheoryElement;

struct AVLNode
{
	signed char balance;            // AVL tree balance
	char isHorn;                    // Whether the atom is horn atom
	short value;                    // Value of the atom (true, false, etc.)
	unsigned int atomNum;           //low word of the key
	int symbolNum;                  //high word of the key
	unsigned int hornRuleNum;       // Valid only for horn atoms. Contains
	                                //  the number of non-eliminated ground
	                                //  horn rules that have this
	                                //  atom in the head.
	unsigned int number;            // Number that associated with the
	                                //  atom and appears in the ground
	                                //  rules in place of the atom.
	GroundTheoryElement *firstRule; // First element in the list of rule
	                                //  that contain this atom.
	AVLNode	*left;
	AVLNode	*right;
};


class Binary
{
	MemNode <AVLNode> nodeSrc; // Source of the AVL tree nodes.
	Stack <AVLNode*> stk;      // Stack

		// Auxiliary methods for traversal:
	void rotateLeft(AVLNode *p);
	void rotateRight(AVLNode *p);

	AVLNode *root; // Pointer to the root node

public:
	Binary(); // CONSTRUCTOR
		// Insertion into the tree
	AVLNode *insert(int symbolNum, unsigned int atomNum);
		// Traversal of the tree
	AVLNode *first();
	AVLNode	*next();
};
#endif
