/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: lexan.cpp
 *
 *  CONTENTS:  LEXICAL ANALYZER : FRONT END, PSGRND.
 *              implemntation of the class 'LexicalAnalyzer'
 *
 *  DATE:     April 13, 2004
 ***************************************************************************/

#include <errno.h>
#include <stdio.h>

#include "lexan.h"
#include "y_tab.h"


// INITIALIZE CLASS ARRAYS:
int LexicalAnalyzer::kwTokens[]={ABS, MAX, MIN, MOD, PRED, VAR};
char LexicalAnalyzer::keywords[][MAX_KW_LEN+1]=
			{"abs","max","min","mod","pred","var"};
LTEntry LexicalAnalyzer::LetterTable[256];

unsigned char LexicalAnalyzer::oneSymbolTokens[]=",.;:*/+-<>=()[]{}|_";
int LexicalAnalyzer::oneSymbolTokenNums[]= {COMMA, PERIOD, SEMICOLON,
		COLON, ASTERISK, SLASH, PLUS, MINUS, LESS, GREATER,
		ASSIGN, LEFTPARENTH, RIGHTPARENTH, LEFTBRACKET,
		RIGHTBRACKET, LEFTBRACE, RIGHTBRACE, OR, UNDERSCORE};

// NOTE: Entries in finalToToken which are not going to be used are
//       indicated as an UNDERSCORE.
int LexicalAnalyzer::finalToToken[STT_FIN_NUM]={UNDERSCORE, UNDERSCORE,
ARROW, COLONMINUS, NOTEQUAL, NOTGREATER, EQUAL, NOTLESS, DOUBLEDOT,
PERIOD, NEWLINE, NEWLINE, NUM, ID, UNDERSCORE};
char LexicalAnalyzer::finalTokens[STT_FIN_NUM][6]={"_", "_", "->", ":-",
"!=", "<=", "==", ">=", "..", ".", "\\n", "\\n", "NUM", "ID", "_"};

OldToken LexicalAnalyzer::oldTokens[OLD_TOKEN_BUF_SIZE];

int LexicalAnalyzer::IDTypes[]=
{ID, ConstantID, VariableID, DataPredID, ProgramPredID, WeightFunID};


/*
int LexicalAnalyzer::scan(int *number)
DESCRIPTION
	THE MAIN METHOD OF THE LEXICAL ANALYZER. Once all files are opened,
	the method might be called to read tokens from the input files
	to be processed by the parser

PARAMETER
	number - Pointer to the integer where the token value will be stored

RETURNS
	Number of token read from the input
*/
int LexicalAnalyzer::scan(int *number)
{
    unsigned char let; // current letter from the input string.
    int num;
    int token, tokenNumber;
    int current=inBufStartToken, state=0;

    while(1)
    {
            // If we are out of input, need to read additional
            //  portion of the input file.
        if(current >= inBufFilled)
        {
                // If the function to read file returns that there is nothing
                //  more to read and if there is no unfinished token,
                // then output the EndMarker and quit.
            if((num=readMoreFromFile(&current)))
            {
                    // For every new file we need to
                    //   initialize the line count:
                if(num == 1) lineCount=1;

                    // no more files => return end marker
                if(num == 2)
                {
                    // Save the last token:
                    oldTokens[curOldToken].fileName=(curFile)?curFile->name:0;
                    oldTokens[curOldToken].lineNum=lineCount;
                    oldTokens[curOldToken].number=0;
                    oldTokens[curOldToken].token=EndMarker;
                    curOldToken=(curOldToken + 1) % OLD_TOKEN_BUF_SIZE;

                    return EndMarker;
                }
            }
        }

            // Finding the next state.
        let=inBuffer[current];

        state=StateTable[state][LetterTable[let].colInd];

        if(state>=STT_HEIGHT)
        {                        // THE NEW STATE IS FINAL:
            state-=STT_HEIGHT;    // Finding the number of the final state
                // Backing up if necessary (to be more precise,
                //   moving forward if necessary)
            current+=FinalTable[state];
            token=finalToToken[state];
            tokenNumber=0;

            if(state<=1) token=LetterTable[inBuffer[current-1]].spec;
            if (state>=10)
            {
                if(state <=11)
                {
                    lineCount++;
                    state=0;
                    inBufStartToken=current;
                    continue;
                }

                    // Mark the end of token.
                let=inBuffer[current];
                inBuffer[current]=0;


                switch(state)
                {
                case 13: // Character string
                        // If the string contains only one underscore, 
                        //  return underscore token number:
                    if(inBuffer[inBufStartToken]=='_' && 
                        inBuffer[inBufStartToken+1]==0)
                    {
                        token=LetterTable[(unsigned)'_'].spec;
                        break; // From the 'switch'
                    }
                        // Checking this string among the keywords:
                    for(num=0;num<ARRAY_SIZE(kwTokens);num++)
                        if(!strcmp(keywords[num],
                                     (char *)(inBuffer+inBufStartToken)))
                        { token=kwTokens[num];  break; }

                        // If we didn't find it, the string is not a keyword.
                        // Therefore, we need to insert it into the symbol table
                    if(num>=ARRAY_SIZE(kwTokens))
                    {
                        tokenNumber=st.insertName(
                                         (char *)(inBuffer+inBufStartToken));
                        num=st.getType(tokenNumber);

                            // Checking whether the type of token will affect
                            // the token number to output:
                        if(num<ARRAY_SIZE(IDTypes)) token=IDTypes[num];
                        else // If token is a numerical constant, we need to
                            // obtain its value and return it to the parser
                            //  instead of a string.
                            if(num == TYPE_NUM_CONST)
                            {
                                token=NUM;
                                tokenNumber=(st.getVal(tokenNumber))->number;
                            }
                            else
                            {
                                fprintf(stderr, "CRITICAL ERROR: Unsupported"
                                    " type of the identifier: %d\n", num);
                                exit(1);
                            }
                    }
                    break;
                case 12:    // Integer
                    tokenNumber=atoi((char *)(inBuffer+inBufStartToken));
                    break;
                case 14:    // End of illegal string of tokens:
                    num = error((char *)(inBuffer+inBufStartToken));
                        // Restore the letter in the input buffer
                    inBuffer[current]=let;
                    inBufStartToken=current;
                    state=0;
                    if(!num)
                    {
                            // Invalidate all the remaining input
                        interruptInput();
                        current=inBufStartToken;    // Initialize current
                    }
                    continue;    // Continue the loop;
                                //  doesn't relate to the switch.
                }

                    // Restore the letter in the input buffer
                inBuffer[current]=let;
            }
            state=0;        // Moving to the initial state.

            *number=tokenNumber;
            inBufStartToken=current;

                // Saving the current token:
            oldTokens[curOldToken].fileName=(curFile)?curFile->name:0;
            oldTokens[curOldToken].lineNum=lineCount;
            oldTokens[curOldToken].number=tokenNumber;
            oldTokens[curOldToken].token=token;
            curOldToken=(curOldToken + 1) % OLD_TOKEN_BUF_SIZE;

            return token;
        }
        else
            current++;    // THE NEW STATE IS NOT FINAL:

            // States 0-1 indicate that the current letter doesn't belong to
            //  any token (i.e this is a space, letter inside a comment, etc.)
        if(state<=1) inBufStartToken=current;
    }
}


// Destructor
LexicalAnalyzer::~LexicalAnalyzer()
{
		// Close all input files:
	closeAllFiles();
		// Release the input buffer if necessary:
	if(inBuffer) free(inBuffer);
	inBuffer=0;
}

// Constructor. NEEDS A SYMBOL TABLE AS A PARAMETER.
LexicalAnalyzer::LexicalAnalyzer(SymbolTable &symbolTable):
				st(symbolTable),EndMarker(ENDMARKER)
{
	int i;
		// Allocate the file input buffer:
	inBuffer=(unsigned char *)malloc(FILE_BUFFER_SIZE);
	if(!inBuffer)
	{
		fprintf(stderr, "CRITICAL ERROR: cannot allocate memory while\n"
		    "constructing the object of the class LexicalAnalyzer\n\n");
		exit(1);
	}
		// Initialize input-managing variables:
	inBufStartToken=inBufFilled=0;
	isInputInterrupt=0;

		// Initialize the file-handling variables:
	firstFile=curFile=lastFile=0;

		// Initialize other variables:
	errorCount=0;
	errorStream=stderr;
	lineCount=-1;
	curOldToken=0;

		// Initialize the array of old tokens (so trying to get token
		//  that was not yet input will not cause unexpected values).
	for(i=0;i<ARRAY_SIZE(oldTokens);i++)
	{
		oldTokens[i].token=EndMarker;
		oldTokens[i].lineNum=-1;
		oldTokens[i].fileName=0;
		oldTokens[i].number=0;
	}

	InitLetterTable(); // Initialize the letter table.
}


/*
int LexicalAnalyzer::addFile(char *FileName)
	This function should be called at least once to start lexical analyzis.
	If the file can be opened, it is added to the list of files to
	 be processed.

PARAMETERS:
	fileName - Name of the file to be opened.
RETURNS:  0     - Operation completed successfully.
          Not 0 - Error: returns value of the variable errno, containing error
          code (the error codes are specified in the standard library)
*/
int LexicalAnalyzer::addFile(char *fileName)
{
	int file;
	InputFile *newNode;

		// Try to open the given file
	file=open(fileName, O_BINARY|O_RDONLY);
	
		// If the file cannot be opened, return the error code:
	if(file==-1) return errno;

	// If the file can be opened, add it to the list of files to be
	//  processed by the lexical analyzer

		// Allocate memory for the new node and name of the file:
	newNode=(InputFile *)malloc(sizeof(InputFile) + strlen(fileName)+1);

	if(!newNode)
	{
		fprintf(stderr, "CRITICAL ERROR: cannot allocate memory"
				"in LexicalAnalyzer::addFile\n");
		exit(1);
	}

		// Initialize the fields of the node:
	newNode->name=(char *)(newNode+1); // name should point to the place
	                                   // right after the node
	strcpy(newNode->name, fileName);
	newNode->file=file;
	newNode->isNew=1;
	newNode->next=0;

		// Append the node to the end of the list:
	if(!lastFile) {curFile=firstFile=lastFile=newNode;}
	else lastFile=lastFile->next=newNode;

		
		// Tell that the operation is completed successfully:
	return 0;
}

/*
int LexicalAnalyzer::readMoreFromFile(int *current)

...
RETURNS
	0 - The current file continues.
	1 - Started reading the new file.
	2 - The last file is finished and there is nothing else to read.
*/

#define WRITE_NEWLINE_TO_INPUT_BUFFER \
	{inBuffer[inBufFilled]='\n'; inBufFilled++;}

int LexicalAnalyzer::readMoreFromFile(int *current)
{
	int toReturn;

		// Checking the length of the current identifier:
		//  If the current token takes a half of the input buffer,
		//   then something must be wrong with the file.
		//  Probably, the file was binary.
	if(*current - inBufStartToken >= FILE_BUFFER_SIZE /2)
	{
		fprintf(errorStream, "\nERROR in the input file: \'%s\'\n"
		     "The program found very long identifier (length == %d)\n"
		     "Probably the file is binary\n"
		     "No more files will be processed and no output will be "
		     "generated\n\n", curFile->name, *current-inBufStartToken);

			// Close all the files.
		closeAllFiles();

		inBufStartToken=*current=inBufFilled = 0;
		WRITE_NEWLINE_TO_INPUT_BUFFER;

			// Return that there is no more information to read.
		return 2;
	}

		// Find the new value of the variable 'current' which is equal to
		// the length of the token that is currently parsed:
	*current-=inBufStartToken;

		// Copy the currently parsed token to the
		//   beginning of the input buffer:
	memcpy(inBuffer, inBuffer + inBufStartToken, *current);
		// Now the last token starts from the beginning of the buffer
		//  and *current points at the first unoccupied place
	inBufStartToken=0;
	inBufFilled=*current;

		// If the current file is finished and closed,
		//   go to the next file:
	if(curFile && curFile->file == -1) curFile=curFile->next;

		// curFile==0 <=> There are no more files to read;
		//  therefore, need to Write the endline and return
		// that there is no more information to read.
	if(!curFile)
	{
		WRITE_NEWLINE_TO_INPUT_BUFFER;
		return 2;
	}

		// If we are here, the current file is not closed;
		//  therefore, we can read from it.
	inBufFilled+=read(curFile->file, inBuffer+inBufFilled,
		FILE_BUFFER_SIZE-inBufFilled);

		// Saving previous status of the file regarding whether it
		// was read from before and mark the file as started:
	toReturn=curFile->isNew;
	curFile->isNew=0;

		// If the function reads less than requested,
		//  we reached end of file. In this case close
		//  the file and put '\n' at the end of input:
	if(inBufFilled < FILE_BUFFER_SIZE)
	{
		close(curFile->file);
		curFile->file=-1;

		WRITE_NEWLINE_TO_INPUT_BUFFER;
	}

		// return value indicates whether a new file has been started.
	return toReturn;
}


/*
void LexicalAnalyzer::interruptInput()
DESCRIPTION
	The method discards all remaining input so that lexical analyzer
	will return the EndMarker. This can be used to stop parsing in 
	case of errors.
*/
void LexicalAnalyzer::interruptInput()
{
	curFile=0;
	inBufStartToken=inBufFilled=0;
	isInputInterrupt=1;
}


/*
void LexicalAnalyzer::closeAllFiles()
DESCRIPTION
	This method closes all the files opened by the calls of
	LexicalAnalyzer::addFile and releases memory allocated to
	store the information regarding these files
*/
void LexicalAnalyzer::closeAllFiles()
{
	for( ; firstFile ; curFile=firstFile, firstFile=firstFile->next,
			free(curFile))
		if(firstFile->file != -1) close(firstFile->file);

	curFile=lastFile=0;
}


/*
int LexicalAnalyzer::getLineNum(int offset, char **fileName)
DESCRIPTION
	Get the line number of the last token returned by the lexical
	analyzer or the token that preceded it. The number of tokens the
	method can look back is specified by the
	constant: OLD_TOKEN_BUF_SIZE.

PARAMETERS
	offset    -  Offset to the token we need to get the line number for.
	              0 corresponds to the last token, 1 to the token
	              before the last, etc.
	fileName  - Place where to put the pointer to the file name
	              corresponding to that line number. This parameter
	              can be NULL.
NOTE
	Dereferencing that pointer is NOT ALWAYS POSSIBLE. Can dereference
	only if the method: 'isFileNameValid' returns TRUE.

RETURNS
	The requested line number
*/
int LexicalAnalyzer::getLineNum(int offset, char **fileName)
{
	int index;

	if(offset >= OLD_TOKEN_BUF_SIZE)
	{
		if(fileName) *fileName=0;
		return -1;
	}

	index = (curOldToken-1-offset+OLD_TOKEN_BUF_SIZE)%OLD_TOKEN_BUF_SIZE;

	if(fileName) *fileName=oldTokens[index].fileName;
	return(oldTokens[index].lineNum);
}


/*
int LexicalAnalyzer::getLastToken(int *number)
DESCRIPTION
	Return the last token scanned by the lexical analyzer

PARAMETERS
	number  -   Pointer to the place that receives the token value

RETURNS
	Number of the last token
*/
int LexicalAnalyzer::getLastToken(int *number)
{
	int i = (curOldToken - 1 + OLD_TOKEN_BUF_SIZE)%OLD_TOKEN_BUF_SIZE;
	if(number) *number=oldTokens[i].number;
	return oldTokens[i].token;
}



/*
void LexicalAnalyzer::reportCurFileName(FILE *file)
DESCRIPTION
	This method prints to "file" the name of the current input file.
	If this method is called several times for the same file, it prints
	the name only once during the first call and then does nothing.
*/
void LexicalAnalyzer::reportCurrentFileName(FILE *file)
{
	static char *oldName=0;

	if(curFile && oldName!=curFile->name)
	{
		fprintf(file, "IN FILE \"%s\":\n", curFile->name);
		oldName=curFile->name;
	}
}


/*
const char *LexicalAnalyzer::tokenToString(
			int token, int number, int genericLevel)
DESCRIPTION
	This method returns the name of the given token.
	It may be used for debugging purposes
*/
const char *LexicalAnalyzer::tokenToString(int token, int number,
		int genericLevel)
{
	static char TmpStr[50];
	int i;
	int isID;

		// Check whether token is an ID:
	for(i=0,isID=0;i<ARRAY_SIZE(IDTypes);i++)
		if(IDTypes[i]==token) {isID=1; break;}

		// Check whether the token is a valid terminal:
	if(token<MinTerminal||token>MaxTerminal)
	{
		sprintf(TmpStr, "Invalid token Number: %d", token);
		return(TmpStr);
	}

	if(!(genericLevel&PR_ID))
		if(isID) return(st.getName(number));

	if(!(genericLevel&PR_NUM))
	{
		if(token==NUM)
		{
			sprintf(TmpStr, "%d", number);
			return(TmpStr);
		}
	}


	for(i=0;i<ARRAY_SIZE(oneSymbolTokenNums);i++)
		if(token==oneSymbolTokenNums[i])
		{
			TmpStr[0]=oneSymbolTokens[i];
			TmpStr[1]=0;
			return(TmpStr);
		}


	for(i=0;i<ARRAY_SIZE(kwTokens);i++)
		if(token==kwTokens[i]) return(keywords[i]);

	for(i=0;i<ARRAY_SIZE(finalToToken); i++)
		if(token==finalToToken[i]) return(finalTokens[i]);

	if(token==EndMarker) return ("%END_MARKER%");

	return("%UNKNOWN%");
}


#define LT(s) LetterTable[s].colInd

/*
void LexicalAnalyzer::InitLetterTable()
DESCRIPTION
	This method is a part of the constructor and can be used to
	initialize the letter table.
*/
void LexicalAnalyzer::InitLetterTable()
{
	int i;
		// Lists of digits, letters,
		//  symbols for the state table, and operators.
	static unsigned char digits[]="0123456789";
	static unsigned char letters[]="ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	static unsigned char stateTableHead[]="  %!.:=<>-\n\r";
	static unsigned char delimiters[]=" \t\000";

		// All symbols, for which table wouldn't be changed after
		//  this 'for' loop, would be treated as illegal
		//  punctuation ('ilp' column)
	int ErrPunct=ARRAY_SIZE(stateTableHead)+1;
	for(i=0;i<256;i++)
	{
		LT(i)=ErrPunct;
		LetterTable[i].spec=0;
	}

		// Check for consistency of the tables:
	if(ARRAY_SIZE(oneSymbolTokens)-1!=ARRAY_SIZE(oneSymbolTokenNums))
	{
		printf("CRITICAL ERROR: Incorrect data in "
			"function LexicalAnalyzer::InitLetterTable\n");
		exit(1);
	}

		// Check for consistency of the tables:
	if(ARRAY_SIZE(stateTableHead)+2!=STT_WIDTH)
	{
		printf("CRITICAL ERROR: Incorrect Header for "
			"the state table or its width \n");
		exit(1);
	}

		// All simple operators, which wouldn't be changed later,
		//  considered as a general punctuation ('lp' column).
		//  These are the only operators requiring initialization of
		//  the 'spec' field.
	for(i=0;i<ARRAY_SIZE(oneSymbolTokens)-1;i++)
	{
		LT(oneSymbolTokens[i])=ARRAY_SIZE(stateTableHead);
		LetterTable[oneSymbolTokens[i]].spec=oneSymbolTokenNums[i];
	}

		// Initializing the table for digits and letters
		//  letters: column 0, digits: column 1.
	for(i=0;i<ARRAY_SIZE(digits)-1;i++) LT(digits[i])=1;
	for(i=0;i<ARRAY_SIZE(letters)-1;i++)
		{ LT(letters[i])=0; LT(letters[i]+'a'-'A')=0;}
	LT((unsigned)'_')=0; // Underscore is treated as a letter.

		// Initialize the letter table for symbols
		//  Each symbol corresponds to a column from 2 to 19.
	for(i=2;i<ARRAY_SIZE(stateTableHead)-1;i++) LT(stateTableHead[i])=i;

		// Initialize the letter table for delimeters:
	for(i=0;i<ARRAY_SIZE(delimiters)-1;i++)
		LT(delimiters[i])=ARRAY_SIZE(stateTableHead)-1;


		// Initialize token number description variables.
	MinTerminal=MaxTerminal=EndMarker;
	for(i=0;i<ARRAY_SIZE(oneSymbolTokenNums);i++)
	{
		if(oneSymbolTokenNums[i]>MaxTerminal)
			MaxTerminal=oneSymbolTokenNums[i];
		if(oneSymbolTokenNums[i]<MinTerminal)
			MinTerminal=oneSymbolTokenNums[i];
	}

	for(i=0;i<ARRAY_SIZE(kwTokens);i++)
	{
		if(kwTokens[i]>MaxTerminal) MaxTerminal=kwTokens[i];
		if(kwTokens[i]<MinTerminal) MinTerminal=kwTokens[i];
	}

	for(i=0;i<ARRAY_SIZE(finalToToken);i++)
	{
		if(finalToToken[i]>MaxTerminal) MaxTerminal=finalToToken[i];
		if(finalToToken[i]<MinTerminal) MinTerminal=finalToToken[i];
	}
}
#undef LT


/*
int LexicalAnalyzer::getTerminalRange(int *min, int *max)
DESCRIPTION
	Get the range for the possible number of tokens.
*/
int LexicalAnalyzer::getTerminalRange(int *min, int *max)
{
	if(min) *min=MinTerminal;
	if(max) *max=MaxTerminal;
	return(MaxTerminal-MinTerminal+1);
}

/*
int LexicalAnalyzer::error(char *cause)
DESCRIPTION
	This method prints an error message. Should be called when the scanner
	detected an illegal token. This method also decides whether the scanner
	should continue scanning the input file or quit.

PARAMETERS
	cause  -    The name of the token that caused an error.

RETURNS
	1  -  The scanner may continue scanning tokens
	0  -  The scanner should close all the files and quit
*/
int LexicalAnalyzer::error(char *cause)
{
	reportCurrentFileName(errorStream);
	fprintf(errorStream, "LEXICAL ERROR");

	if(lineCount != -1) fprintf(errorStream, " in line %d: ", lineCount);
	else fprintf(errorStream, ": ");
	fprintf(errorStream, "Illegal token: \'%s\'\n", cause);

	errorCount++;
	if(errorCount > MAX_LEXICAL_ERRORS)
	{
		fprintf(errorStream, "TOO MANY LEXICAL ERRORS "
			"Scanning is stopped; no output will be produced\n");
		return 0;
	}

	return 1;
}

// STATE TABLE:

/*
Table Notation (Columns are numbered starting from 0):
0 - 1 Columns:
      L    -  Any letter or '_'
      D    -  Any digit

2 - 8 Columns Transition by the corresponding symbol
9-12 Columns:
      \n   -  Endline symbol: '\n'.
      \r   -  Carret return symbol: '\r'.
      sp   -  Space, TAB.
      lp   -  "Legal punctuation". Any token which has the following properties:
                  a) It is represented by only one character.
                  b) This character doesn't occur in any other token
      ilp  -  "Illegal punctuation". No token contains this symbol.
*/

#define F STT_HEIGHT
char LexicalAnalyzer::StateTable[STT_HEIGHT][STT_WIDTH]=
{
//        L    D    %    !    .    :    =    <    >    -    \n   \r   sp   lp   ilp
/* 0 */  {3,   11,  1,   4,   8,   5,   10,  6,   7,   9,   F+11,12  ,0,   F+0 ,2   },
/* 1 */  {1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   F+11,12  ,1,   1,   1   },
/* 2 */  {F+14,F+14,F+14,F+14,F+14,F+14,F+14,F+14,F+14,F+14,F+14,F+14,F+14,F+14,2   },
/* 3 */  {3,   3,   F+13,F+13,F+13,F+13,F+13,F+13,F+13,F+13,F+13,F+13,F+13,F+13,F+13},
/* 4 */  {F+14,F+14,F+14,F+14,F+14,F+14,F+4, F+14,F+14,F+14,F+14,F+14,F+14,F+14,2   },
/* 5 */  {F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+3, F+1 ,F+1 ,F+1 ,F+1 ,F+1 },
/* 6 */  {F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+5, F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 },
/* 7 */  {F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+7, F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 },
/* 8 */  {F+9 ,F+9 ,F+9 ,F+9 ,F+8, F+9 ,F+9 ,F+9 ,F+9 ,F+9 ,F+9 ,F+9 ,F+9 ,F+9 ,F+9 },
/* 9 */  {F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+2 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 },
/*10 */  {F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+6, F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 ,F+1 },
/*11 */  {F+12,11,  F+12,F+12,F+12,F+12,F+12,F+12,F+12,F+12,F+12,F+12,F+12,F+12,F+12},
/*12 */  {F+10,F+10,F+10,F+10,F+10,F+10,F+10,F+10,F+10,F+10,F+11,F+10,F+10,F+10,F+10}
};
#undef F
/*
THE LIST OF NON-FINAL STATES:

0  Starting state
1  Inside a comment
2  In the string of illegal tokens.
3  In word
4  Found !
5  Found :
6  Found <
7  Found >
8  Found .
9  Found -
10 Found =
11 In number
12 Found \r


THE LIST OF FINAL STATES:

F+0  (N)Found one-character token
F+1  (Y)Found one-character token
F+2  (N)Found Token ->
F+3  (N)Found Token :-
F+4  (N)Token !=
F+5  (N)Token <=
F+6  (N)Token ==
F+7  (N)Token >=
F+8  (N)Found Token ..
F+9  (Y)Found token .
F+10 (Y)End of Line. Increate the line count.
F+11 (N)End of Line. Increase the line count.
F+12 (Y)End of Integer Number
F+13 (Y)End of Identifier
F+14 (Y)End of an illegal token
*/
// This table tells the lexical analyzer whether it needs to backtrack
// 0: YES, 1: NO.
char LexicalAnalyzer::FinalTable[STT_FIN_NUM]={1,0,1,1,1,1,1,1,1,0,0,1,0,0,0};

/*
Diagram: Implementation of the Lexical Analyzer:


  [INPUT STRING]                           Current State
      |                                            |
      \                                            \       /- Not Final State...
1)...letter->[LetterTable.colInd]--Column #-> [StateTable]<
                                                           \- Final State->



                    /-Incorrect token...
                  /---Word -\---------- Identifier ----------------> Token # ...
2)->Final State -<            \------- Keyword -----[kwTokens] ----> Token # ...
         |        \-------------Specific Token------[finalToToken]-> Token # ...
         |          \-General Token
         |                 \
         |             ...  letter-->[LetterTable.spec]------------> Token # ...
         |
          \
     [FinalTable] -Whether to back up->....

NOTE: The lexical analyzer outputs TOKEN NUMBERS. In most cases, they are
        equal to the number of the final state (which means difference between
		table number of the given final state and the table number
		of the first final state.
*/
